/*
 * Copyright (c) 2012 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#if defined(__linux__) || defined(__MINT__)
# define _GNU_SOURCE /* strptime(), getsubopt() */
#endif

#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "extern.h"

#define	UNCONST(a)	((void *)(uintptr_t)(const void *)(a))

struct	c {
	const struct expr	 *last;
	const struct sym	**domains;
	size_t			  domainsz;
	const char		 *func;
	int			  flags;
#define	C_FRAGMENT		 (0x01)
};

static	void	 expr(const struct expr *, int);
static	void	 expr_value(const struct expr_value *);
static	void	 expr_binary(const struct expr_binary *);
static	void	 expr_unary(const struct expr_unary *);

static	void	 expr_c_close(void *);
static	void	 expr_c_end(const struct expr *, void *);
static	void	 expr_c_free(void *);
static	void	 expr_c_start(const struct sym *, void *);
static	void	 expr_c_open(void *);
static	int	 expr_c_options(struct c *, char *);
static	void	 expr_c_write(enum rule, const struct expr *, void *);

static	const char * const unaries[UT__MAX] = {
	"-", /* UT_NEG */
	"log", /* UT_LN */
	"sin", /* UT_SIN */
	"cos", /* UT_COS */
	"tan", /* UT_TAN */
	"acos", /* UT_ARCCOS */
	"asin", /* UT_ARCSIN */
	"atan", /* UT_ARCTAN */
	"???", /* UT_INTEGRAL */
	"erf" /* UT_ERF */
};

static	const char * const binaries[BT__MAX] = {
	"+", /* BT_ADD */
	"-", /* BT_SUB */
	"*", /* BT_MUL */
	"/", /* BT_DIV */
	NULL /* BT_EXP */
};

/* ARGSUSED */
static void
expr_c_write(enum rule r, const struct expr *ep, void *arg)
{

	/* Do nothing. */
}

/* ARGSUSED */
static void
expr_c_open(void *arg)
{
	struct c	*p = arg;
	
	if (C_FRAGMENT & p->flags)
		return;

	puts("#include <math.h>");
	puts("#include <stdlib.h>");
	puts("");
}

/* ARGSUSED */
static void
expr_c_start(const struct sym *sym, void *arg)
{

	/* Do nothing. */
}

/* ARGSUSED */
static void
expr_c_end(const struct expr *ep, void *arg)
{
	struct c	*p;

	p = (struct c *)arg;
	p->last = ep;
}

/* ARGSUSED */
static void
expr_c_close(void *arg)
{
	struct c	*p;
	size_t		 i;

	p = (struct c *)arg;
	assert(p->last);

	puts("double");
	printf("%s(", p->func);
	for (i = 0; i < p->domainsz; i++) {
		if (i)
			putchar(',');
		printf("double %s", p->domains[i]->name);
	}
	puts(")");
	puts("{");
	printf("return");
	expr(p->last, 1);
	puts(";");
	puts("}");
}

/* ARGSUSED */
static void
expr_c_free(void *arg)
{

	free(arg);
}

/* ARGSUSED */
struct writer *
expr_write_c(char *cp, const char *func,
		const struct sym **domains, size_t domainsz)
{
	struct writer	*p;
	struct c	*arg;

	arg = calloc(1, sizeof(struct c));
	if (NULL == arg) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}
	arg->domains = domains;
	arg->domainsz = domainsz;
	arg->func = func;

	p = calloc(1, sizeof(struct writer));
	if (NULL == p) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	p->write = expr_c_write;
	p->write_close = expr_c_close;
	p->write_end = expr_c_end;
	p->write_open = expr_c_open;
	p->write_start = expr_c_start;
	p->write_free = expr_c_free;
	p->arg = arg;

	if ( ! expr_c_options(arg, cp)) {
		free(arg);
		free(p);
		return(NULL);
	}

	return(p);
}

static void
expr(const struct expr *ep, int parens)
{

	if (parens)
		putchar('(');

	switch (ep->type) {
	case (ET_BINARY):
		expr_binary(&ep->val.bin);
		break;
	case (ET_UNARY):
		expr_unary(&ep->val.un);
		break;
	case (ET_VALUE):
		expr_value(&ep->val.val);
		break;
	default:
		abort();
		/* NOTREACHED */
	}

	if (parens)
		putchar(')');
}

static void
expr_unary(const struct expr_unary *ep)
{

	assert(ep->type < UT__MAX);
	printf("%s", unaries[ep->type]);
	expr(ep->val, 1);
}

static void
expr_binary(const struct expr_binary *ep)
{
	struct expr	*lhs, *rhs;

	lhs = ep->lhs;
	rhs = ep->rhs;

	/*
	 * Gnuplot doesn't recognise Euler's number as a constant, so
	 * instead of using e**x we use the exp() function.
	 */

	if (BT_EXP == ep->type) {
		if (ET_VALUE == lhs->type && 
				VT_IDENT == lhs->val.val.type && 
				0 == strcmp("e", lhs->val.val.val.sym->name)) {
			printf("exp");
			expr(ep->rhs, 1);
			return;
		}
		if (ET_VALUE == rhs->type &&
				VT_NUM == rhs->val.val.type &&
				0.5L == rhs->val.val.val.num) {
			printf("sqrt");
			expr(ep->lhs, 1);
			return;
		}
		printf("pow(");
		expr(ep->lhs, 0);
		printf(",");
		expr(ep->rhs, 0);
		printf(")");
		return;
	}

	expr(ep->lhs, 1);
	printf("%s", binaries[ep->type]);
	expr(ep->rhs, 1);
}

static void
expr_value(const struct expr_value *ep)
{

	switch (ep->type) {
	case (VT_NUM):
		printf("%Lg", ep->val.num);
		break;
	case (VT_IDENT):
		if (0 == strcmp("pi", ep->val.sym->name))
			printf("M_PI");
		else
			printf("%s", ep->val.sym->name);
		break;
	default:
		abort();
		/* NOTREACHED */
	}
}

static int
expr_c_options(struct c *p, char *arg)
{
	char		*v, *o;
	const char	*toks[2]; 

	toks[0] = "fragment";
	toks[1] = NULL;

	if (NULL == arg)
		return(1);

	while (*arg) {
		o = arg;
		switch (getsubopt(&arg, UNCONST(toks), &v)) {
		case (0):
			p->flags |= C_FRAGMENT;
			break;
		default:
			fprintf(stderr, "-O%s: Bad argument\n", o);
			return(0);
		}
	}

	return(1);
}
