/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <stdlib.h>

#include "extern.h"

struct	diff {
	struct writer		*write;
	struct symtab		*tab;
	const struct sym	*sym;
	const struct expr	*top;
	int			 order;
};

#define	WRITE(e, d, r) \
	do { \
		(e)->flags |= EXPR_DIFF; \
		expr_write((d)->write, (r), (d)->top); \
		(e)->flags &= ~EXPR_DIFF; \
	} while (0 /* CONSTCOND */)

static	int	 expr_constness(const struct sym *, struct expr *);
static	void	 expr_diff(const struct diff *, struct expr *);
static	void	 expr_diff_add(const struct diff *, struct expr *);
static	void	 expr_diff_arccos(const struct diff *, struct expr *);
static	void	 expr_diff_arcsin(const struct diff *, struct expr *);
static	void	 expr_diff_arctan(const struct diff *, struct expr *);
static	void	 expr_diff_bin(const struct diff *, struct expr *);
static	void	 expr_diff_cos(const struct diff *, struct expr *);
static	void	 expr_diff_div(const struct diff *, struct expr *);
static	void	 expr_diff_erf(const struct diff *, struct expr *);
static	void	 expr_diff_exp(const struct diff *, struct expr *);
static	void	 expr_diff_integral(const struct diff *, struct expr *);
static	void	 expr_diff_ln(const struct diff *, struct expr *);
static	void	 expr_diff_mul(const struct diff *, struct expr *);
static	void	 expr_diff_neg(const struct diff *, struct expr *);
static	void	 expr_diff_sin(const struct diff *, struct expr *);
static	void	 expr_diff_tan(const struct diff *, struct expr *);
static	void	 expr_diff_un(const struct diff *, struct expr *);
static	void	 expr_diff_value(const struct diff *, struct expr *);
static	void	 expr_switch(struct expr *, 
			const struct expr *, const struct sym *);
static	int	 expr_switch_check(struct expr **, 
			const struct expr *, const struct sym *);
static	void	 expr_reset(struct expr *);

typedef	void	(*diff)(const struct diff *, struct expr *);

static	const	 diff expr_diffs[ET__MAX] = {
	expr_diff_value, /* ET_VALUE */
	expr_diff_un, /* ET_UNARY */
	expr_diff_bin /* ET_BINARY */
};

static	const	 diff expr_bin_diffs[BT__MAX] = {
	expr_diff_add, /* BT_ADD */
	expr_diff_add, /* BT_SUB */
	expr_diff_mul, /* BT_MUL */
	expr_diff_div, /* BT_DIV */
	expr_diff_exp /* BT_EXP */
};

static	const	 diff expr_un_diffs[UT__MAX] = {
	expr_diff_neg, /* UT_NEG */
	expr_diff_ln, /* UT_LN */
	expr_diff_sin, /* UT_SIN */
	expr_diff_cos, /* UT_COS */
	expr_diff_tan, /* UT_TAN */
	expr_diff_arccos, /* UT_ARCCOS */
	expr_diff_arcsin, /* UT_ARCSIN */
	expr_diff_arctan, /* UT_ARCTAN */
	expr_diff_integral, /* UT_INTEGRAL */
	expr_diff_erf /* UT_ERF */
};

static void
expr_reset(struct expr *ep)
{
	size_t	 	 i;

	switch (ep->type) {
	case (ET_BINARY):
		expr_reset(ep->val.bin.lhs);
		expr_reset(ep->val.bin.rhs);
		break;
	case (ET_UNARY):
		expr_reset(ep->val.un.val);
		for (i = 0; i < ep->val.un.parmsz; i++)
			expr_reset(ep->val.un.parms[i]);
		break;
	default:
		break;
	}

	ep->flags = 0;
}

/*
 * Calculate whether an expression is constant: all of its
 * subexpressions consist only of constant operations, i.e., those on
 * numbers or non-variable symbols.  Returns non-zero if the expression
 * has been set as a constant.
 *
 * FIXME: "fl" can overflow.
 */
static int
expr_constness(const struct sym *sp, struct expr *ep)
{
	int		 fl;
	struct sym	*sym;

	fl = 0;

	switch (ep->type) {
	case (ET_BINARY):
		fl += expr_constness(sp, ep->val.bin.lhs);
		fl += expr_constness(sp, ep->val.bin.rhs);
		break;
	case (ET_UNARY):
		if (UT_INTEGRAL != ep->val.un.type) {
			fl += expr_constness(sp, ep->val.un.val);
			break;
		}

		/*
		 * Special rules for integrals.  If the minimum and
		 * maximum are constant with respect to the symbol of
		 * differentiation, the integral is constant...
		 */
		fl += expr_constness(sp, ep->val.un.parms[0]);
		fl += expr_constness(sp, ep->val.un.parms[1]);
		if (0 == fl)
			break;

		/*
		 * If they aren't, see whether the integral non-constant
		 * with respect to the variable of integration.
		 */

		assert(ET_VALUE == ep->val.un.parms[2]->type);
		assert(VT_IDENT == ep->val.un.parms[2]->val.val.type);
		sym = ep->val.un.parms[2]->val.val.val.sym;
		if ( ! expr_constness(sym, ep->val.un.val))
			fl = 0;
		break;
	case (ET_VALUE):
		if (VT_IDENT != ep->val.val.type)
			break;
		if (sp == ep->val.val.val.sym)
			fl++;
		break;
	default:
		break;
	}

	ep->flags &= ~EXPR_CONST;
	ep->flags |= fl ? 0 : EXPR_CONST;
	return(fl);
}

static void
expr_diff_neg(const struct diff *dp, struct expr *ep)
{

	assert( ! (EXPR_CONST & ep->flags));

	WRITE(ep, dp, RULE_COEFF);

	expr_diff(dp, ep->val.un.val);
}

static void
expr_diff_add(const struct diff *dp, struct expr *ep)
{

	assert( ! (EXPR_CONST & ep->flags));

	WRITE(ep, dp, RULE_SUM);

	expr_diff(dp, ep->val.bin.lhs);
	expr_diff(dp, ep->val.bin.rhs);
}

static void
expr_diff_mul(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs, *rhs;

	assert( ! (EXPR_CONST & ep->flags));

	WRITE(ep, dp, RULE_PROD);

	lhs = ep->val.bin.lhs;
	rhs = ep->val.bin.rhs;

	if (EXPR_CONST & lhs->flags || EXPR_CONST & rhs->flags) {
		if (EXPR_CONST & lhs->flags)
			expr_diff(dp, rhs);
		else
			expr_diff(dp, lhs);
		return;
	}

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_ADD;
	ep->val.bin.lhs = expr_alloc(ET_BINARY);
	ep->val.bin.lhs->val.bin.type = BT_MUL;
	ep->val.bin.lhs->val.bin.lhs = lhs;
	ep->val.bin.lhs->val.bin.rhs = expr_clone(rhs);
	ep->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.type = BT_MUL;
	ep->val.bin.rhs->val.bin.lhs = expr_clone(lhs);
	ep->val.bin.rhs->val.bin.rhs = rhs;

	expr_diff(dp, lhs);
	expr_diff(dp, rhs);
}

static void
expr_diff_div(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs, *rhs, *new, *new2;

	assert( ! (EXPR_CONST & ep->flags));

	lhs = ep->val.bin.lhs;
	rhs = ep->val.bin.rhs;

	if (EXPR_CONST & lhs->flags) {
		WRITE(ep, dp, RULE_RECIP);

		expr_zero(ep);

		ep->type = ET_BINARY;
		ep->val.bin.type = BT_MUL;
		ep->val.bin.lhs = lhs;
		ep->val.bin.rhs = new = expr_alloc(ET_UNARY);
		new->val.un.type = UT_NEG;
		new->val.un.val = new2 = expr_alloc(ET_BINARY);
		new2->val.bin.type = BT_DIV;
		new2->val.bin.lhs = rhs;
		new2->val.bin.rhs = expr_alloc(ET_BINARY);
		new2->val.bin.rhs->val.bin.type = BT_EXP;
		new2->val.bin.rhs->val.bin.lhs = expr_clone(rhs);
		new2->val.bin.rhs->val.bin.rhs = expr_alloc(ET_VALUE);
		new2->val.bin.rhs->val.bin.rhs->val.val.type = VT_NUM;
		new2->val.bin.rhs->val.bin.rhs->val.val.val.num = 2.0L;
		expr_diff(dp, rhs);
		return;
	} else if (EXPR_CONST & rhs->flags) {
		WRITE(ep, dp, RULE_COEFF);
		expr_diff(dp, lhs);
		return;
	}

	WRITE(ep, dp, RULE_QUOT);

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_DIV;
	ep->val.bin.lhs = expr_alloc(ET_BINARY);
	ep->val.bin.lhs->val.bin.type = BT_SUB;
	ep->val.bin.lhs->val.bin.lhs = expr_alloc(ET_BINARY);
	ep->val.bin.lhs->val.bin.lhs->val.bin.type = BT_MUL;
	ep->val.bin.lhs->val.bin.lhs->val.bin.lhs = lhs;
	ep->val.bin.lhs->val.bin.lhs->val.bin.rhs = expr_clone(rhs);
	ep->val.bin.lhs->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.lhs->val.bin.rhs->val.bin.type = BT_MUL;
	ep->val.bin.lhs->val.bin.rhs->val.bin.lhs = expr_clone(lhs);
	ep->val.bin.lhs->val.bin.rhs->val.bin.rhs = rhs;
	ep->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.type = BT_EXP;
	ep->val.bin.rhs->val.bin.lhs = expr_clone(rhs);
	ep->val.bin.rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.rhs->val.val.type = VT_NUM;
	ep->val.bin.rhs->val.bin.rhs->val.val.val.num = 2.0L;

	expr_diff(dp, lhs);
	expr_diff(dp, rhs);
}

static void
expr_diff_exp(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs, *rhs, *new;

	assert( ! (EXPR_CONST & ep->flags));

	WRITE(ep, dp, RULE_POWER);

	lhs = ep->val.bin.lhs;
	rhs = ep->val.bin.rhs;

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_MUL;
	ep->val.bin.lhs = expr_alloc(ET_BINARY);
	ep->val.bin.lhs->val.bin.type = BT_EXP;
	ep->val.bin.lhs->val.bin.lhs = expr_clone(lhs);
	ep->val.bin.lhs->val.bin.rhs = expr_clone(rhs);
	ep->val.bin.rhs = new = expr_alloc(ET_BINARY);

	new->val.bin.type = BT_ADD;
	new->val.bin.lhs = expr_alloc(ET_BINARY);
	new->val.bin.lhs->val.bin.type = BT_MUL;
	new->val.bin.lhs->val.bin.lhs = lhs;
	new->val.bin.lhs->val.bin.rhs = expr_alloc(ET_BINARY);
	new->val.bin.lhs->val.bin.rhs->val.bin.type = BT_DIV;
	new->val.bin.lhs->val.bin.rhs->val.bin.lhs = expr_clone(rhs);
	new->val.bin.lhs->val.bin.rhs->val.bin.rhs = expr_clone(lhs);
	new->val.bin.rhs = expr_alloc(ET_BINARY);
	new->val.bin.rhs->val.bin.type = BT_MUL;
	new->val.bin.rhs->val.bin.lhs = rhs;
	new->val.bin.rhs->val.bin.rhs = expr_alloc(ET_UNARY);
	new->val.bin.rhs->val.bin.rhs->val.un.type = UT_LN;
	new->val.bin.rhs->val.bin.rhs->val.un.val = expr_clone(lhs);

	expr_diff(dp, lhs);
	expr_diff(dp, rhs);
}

static void
expr_diff_sin(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_TRIG);

	lhs = ep->val.un.val;
	
	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_MUL;
	ep->val.bin.lhs = expr_alloc(ET_UNARY);
	ep->val.bin.lhs->val.un.type = UT_COS;
	ep->val.bin.lhs->val.un.val = expr_clone(lhs);
	ep->val.bin.rhs = lhs;

	expr_diff(dp, lhs);
}

/*
 * If the symbol of a value is what's being integrated over, replace the
 * symbol with "new", which is either the minimum or the maximum
 * depending on who's calling us.
 */
static int
expr_switch_check(struct expr **ep, 
		const struct expr *new, 
		const struct sym *sp)
{

	if (ET_VALUE != (*ep)->type)
		return(0);
	if (VT_IDENT != (*ep)->val.val.type)
		return(0);
	if (sp != (*ep)->val.val.val.sym)
		return(0);

	expr_free(*ep);
	*ep = expr_clone(new);
	return(1);
}

/*
 * Check whether we want to switch the domain of a function's
 * subexpression.  This is used with the fundamental theorem of
 * calculus, when the domain of an integral is replaced by its minimum
 * and maximum.  This calls through to expr_switch_check(), which does
 * the actual switch itself (since we do it by pointers, this passes its
 * subexpression value into the function).
 */
static void
expr_switch(struct expr *ep, 
		const struct expr *new, 
		const struct sym *sp)
{

	switch (ep->type) {
	case (ET_BINARY):
		if ( ! expr_switch_check(&ep->val.bin.lhs, new, sp))
			expr_switch(ep->val.bin.lhs, new, sp);
		if ( ! expr_switch_check(&ep->val.bin.rhs, new, sp))
			expr_switch(ep->val.bin.rhs, new, sp);
		break;
	case (ET_UNARY):
		if ( ! expr_switch_check(&ep->val.un.val, new, sp))
			expr_switch(ep->val.un.val, new, sp);
		break;
	default:
		break;
	}
}

static void
expr_diff_integral(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs, *min, *max, *dx;
	struct sym	*sym;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_CALC);

	assert(3 == ep->val.un.parmsz);

	lhs = ep->val.un.val;
	min = ep->val.un.parms[0];
	max = ep->val.un.parms[1];
	dx = ep->val.un.parms[2];
	assert(ET_VALUE == dx->type);
	assert(VT_IDENT == dx->val.val.type);
	sym = dx->val.val.val.sym;

	ep->val.un.parmsz = 0;
	/*expr_zero(ep);*/
	free(ep->val.un.parms);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_SUB;
	ep->val.bin.lhs = expr_alloc(ET_BINARY);
	ep->val.bin.lhs->val.bin.type = BT_MUL;
	ep->val.bin.lhs->val.bin.lhs = expr_clone(lhs);

	if ( ! expr_switch_check(&ep->val.bin.lhs->val.bin.lhs, max, sym))
		expr_switch(ep->val.bin.lhs->val.bin.lhs, max, sym);

	ep->val.bin.lhs->val.bin.rhs = max;
	ep->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.type = BT_MUL;
	ep->val.bin.rhs->val.bin.lhs = expr_clone(lhs);

	if ( ! expr_switch_check(&ep->val.bin.rhs->val.bin.lhs, min, sym))
		expr_switch(ep->val.bin.rhs->val.bin.lhs, min, sym);

	ep->val.bin.rhs->val.bin.rhs = min;

	expr_free(lhs);
	expr_free(dx);

	expr_diff(dp, max);
	expr_diff(dp, min);
}

static void
expr_diff_erf(const struct diff *dp, struct expr *ep)
{
	struct expr	*val;
	struct sym	*pi, *e;

	pi = symtab_add(dp->tab, "pi", 2);
	e = symtab_add(dp->tab, "e", 1);

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_ERF);

	val = ep->val.un.val;

	expr_zero(ep);
	
	ep->type = ET_BINARY;
	ep->val.bin.type = BT_MUL;
	ep->val.bin.lhs = val;
	ep->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.type = BT_MUL;
	ep->val.bin.rhs->val.bin.lhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.lhs->val.bin.type = BT_DIV;
	ep->val.bin.rhs->val.bin.lhs->val.bin.lhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.lhs->val.bin.lhs->val.val.type = VT_NUM;
	ep->val.bin.rhs->val.bin.lhs->val.bin.lhs->val.val.val.num = 2.0L;
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs->val.bin.type = BT_EXP;
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs->val.bin.lhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs->val.bin.lhs->val.val.type = VT_IDENT;
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs->val.bin.lhs->val.val.val.sym = pi;
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs->val.bin.rhs->val.val.type = VT_NUM;
	ep->val.bin.rhs->val.bin.lhs->val.bin.rhs->val.bin.rhs->val.val.val.num = 0.5L;
	ep->val.bin.rhs->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.rhs->val.bin.type = BT_EXP;
	ep->val.bin.rhs->val.bin.rhs->val.bin.lhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.rhs->val.bin.lhs->val.val.type = VT_IDENT;
	ep->val.bin.rhs->val.bin.rhs->val.bin.lhs->val.val.val.sym = e;
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs = expr_alloc(ET_UNARY);
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.un.type = UT_NEG;
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.un.val = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.un.val->val.bin.type = BT_EXP;
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.un.val->val.bin.lhs = expr_clone(val);
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.un.val->val.bin.rhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.un.val->val.bin.rhs->val.val.type = VT_NUM;
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.un.val->val.bin.rhs->val.val.val.num = 2.0L;

	expr_diff(dp, val);
}

static void
expr_diff_tan(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_TRIG);

	lhs = ep->val.un.val;

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_MUL;
	ep->val.bin.lhs = lhs;
	ep->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.type = BT_ADD;
	ep->val.bin.rhs->val.bin.lhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.lhs->val.val.type = VT_NUM;
	ep->val.bin.rhs->val.bin.lhs->val.val.val.num = 1.0L;
	ep->val.bin.rhs->val.bin.rhs = expr_alloc(ET_BINARY);
	ep->val.bin.rhs->val.bin.rhs->val.bin.type = BT_EXP;
	ep->val.bin.rhs->val.bin.rhs->val.bin.lhs = expr_alloc(ET_UNARY);
	ep->val.bin.rhs->val.bin.rhs->val.bin.lhs->val.un.type = UT_TAN;
	ep->val.bin.rhs->val.bin.rhs->val.bin.lhs->val.un.val = expr_clone(lhs);
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.val.type = VT_NUM;
	ep->val.bin.rhs->val.bin.rhs->val.bin.rhs->val.val.val.num = 2.0L;

	expr_diff(dp, lhs);
}

static void
expr_diff_arccos(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs, *rhs;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_TRIG);

	lhs = ep->val.un.val;

	expr_zero(ep);

	ep->type = ET_UNARY;
	ep->val.un.type = UT_NEG;
	ep->val.un.val = expr_alloc(ET_BINARY);
	rhs = ep->val.un.val;
	rhs->val.bin.type = BT_DIV;
	rhs->val.bin.lhs = lhs;
	rhs->val.bin.rhs = expr_alloc(ET_BINARY);
	rhs = rhs->val.bin.rhs;
	rhs->val.bin.type = BT_EXP;
	rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	rhs->val.bin.rhs->val.val.type = VT_NUM;
	rhs->val.bin.rhs->val.val.val.num = 0.5L;
	rhs->val.bin.lhs = expr_alloc(ET_BINARY);
	rhs = rhs->val.bin.lhs;
	rhs->val.bin.type = BT_SUB;
	rhs->val.bin.lhs = expr_alloc(ET_VALUE);
	rhs->val.bin.lhs->val.val.type = VT_NUM;
	rhs->val.bin.lhs->val.val.val.num = 1.0L;
	rhs->val.bin.rhs = expr_alloc(ET_BINARY);
	rhs->val.bin.rhs->val.bin.type = BT_EXP;
	rhs->val.bin.rhs->val.bin.lhs = expr_clone(lhs);
	rhs->val.bin.rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	rhs->val.bin.rhs->val.bin.rhs->val.val.type = VT_NUM;
	rhs->val.bin.rhs->val.bin.rhs->val.val.val.num = 2.0L;

	expr_diff(dp, lhs);
}

static void
expr_diff_arcsin(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs, *rhs;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_TRIG);

	lhs = ep->val.un.val;

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_DIV;
	ep->val.bin.lhs = lhs;
	ep->val.bin.rhs = expr_alloc(ET_BINARY);
	rhs = ep->val.bin.rhs;
	rhs->val.bin.type = BT_EXP;
	rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	rhs->val.bin.rhs->val.val.type = VT_NUM;
	rhs->val.bin.rhs->val.val.val.num = 0.5L;
	rhs->val.bin.lhs = expr_alloc(ET_BINARY);
	rhs = rhs->val.bin.lhs;
	rhs->val.bin.type = BT_SUB;
	rhs->val.bin.lhs = expr_alloc(ET_VALUE);
	rhs->val.bin.lhs->val.val.type = VT_NUM;
	rhs->val.bin.lhs->val.val.val.num = 1.0L;
	rhs->val.bin.rhs = expr_alloc(ET_BINARY);
	rhs->val.bin.rhs->val.bin.type = BT_EXP;
	rhs->val.bin.rhs->val.bin.lhs = expr_clone(lhs);
	rhs->val.bin.rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	rhs->val.bin.rhs->val.bin.rhs->val.val.type = VT_NUM;
	rhs->val.bin.rhs->val.bin.rhs->val.val.val.num = 2.0L;

	expr_diff(dp, lhs);
}

static void
expr_diff_arctan(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs, *rhs;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_TRIG);

	lhs = ep->val.un.val;

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_DIV;
	ep->val.bin.lhs = lhs;
	ep->val.bin.rhs = rhs = expr_alloc(ET_BINARY);

	rhs->val.bin.type = BT_ADD;
	rhs->val.bin.lhs = expr_alloc(ET_VALUE);
	rhs->val.bin.lhs->val.val.type = VT_NUM;
	rhs->val.bin.lhs->val.val.val.num = 1.0L;
	rhs->val.bin.rhs = expr_alloc(ET_BINARY);
	rhs->val.bin.rhs->val.bin.type = BT_EXP;
	rhs->val.bin.rhs->val.bin.lhs = expr_clone(lhs);
	rhs->val.bin.rhs->val.bin.rhs = expr_alloc(ET_VALUE);
	rhs->val.bin.rhs->val.bin.rhs->val.val.type = VT_NUM;
	rhs->val.bin.rhs->val.bin.rhs->val.val.val.num = 2.0L;

	expr_diff(dp, lhs);
}

static void
expr_diff_cos(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_TRIG);

	lhs = ep->val.un.val;

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_MUL;
	ep->val.bin.lhs = expr_alloc(ET_UNARY);
	ep->val.bin.lhs->val.un.type = UT_NEG;
	ep->val.bin.lhs->val.un.val = expr_alloc(ET_UNARY);
	ep->val.bin.lhs->val.un.val->val.un.type = UT_SIN;
	ep->val.bin.lhs->val.un.val->val.un.val = expr_clone(lhs);
	ep->val.bin.rhs = lhs;

	expr_diff(dp, lhs);
}

static void
expr_diff_ln(const struct diff *dp, struct expr *ep)
{
	struct expr	*lhs;

	assert( ! (EXPR_CONST & ep->flags));
	WRITE(ep, dp, RULE_LN);

	lhs = ep->val.un.val;

	expr_zero(ep);

	ep->type = ET_BINARY;
	ep->val.bin.type = BT_DIV;
	ep->val.bin.lhs = lhs;
	ep->val.bin.rhs = expr_clone(lhs);
	expr_diff(dp, lhs);
}

static void
expr_diff_bin(const struct diff *dp, struct expr *ep)
{

	if ( ! (EXPR_CONST & ep->flags)) {
		(*expr_bin_diffs[ep->val.bin.type])(dp, ep);
		return;
	}

	WRITE(ep, dp, RULE_CONST);

	expr_free(ep->val.bin.lhs);
	expr_free(ep->val.bin.rhs);
	expr_zero(ep);

	ep->type = ET_VALUE;
	ep->val.val.type = VT_NUM;
	ep->val.val.val.num = 0.0L;
}

static void
expr_diff_un(const struct diff *dp, struct expr *ep)
{

	if ( ! (EXPR_CONST & ep->flags)) {
		(*expr_un_diffs[ep->val.un.type])(dp, ep);
		return;
	}

	WRITE(ep, dp, RULE_CONST);

	expr_free(ep->val.un.val);
	expr_zero(ep);

	ep->type = ET_VALUE;
	ep->val.val.type = VT_NUM;
	ep->val.val.val.num = 0.0L;
}

/*
 * Values always follow the constant rule or linear-function rule with a
 * coefficient of one.
 */
static void
expr_diff_value(const struct diff *dp, struct expr *ep)
{

	WRITE(ep, dp, EXPR_CONST & ep->flags ?  RULE_CONST : RULE_COEFF);

	/* Constant/linear-function rule. */

	ep->val.val.type = VT_NUM;
	ep->val.val.val.num = EXPR_CONST & ep->flags ? 0.0L : 1.0L;
}

/*
 * Top-level differentiation of an expression.  Expressions can be
 * binary, unary, or simply a value; drill down into their specific
 * handlers and continue.
 */
static void
expr_diff(const struct diff *dp, struct expr *ep)
{

	assert(*expr_diffs[ep->type]);
	(*expr_diffs[ep->type])(dp, ep);
	expr_reduce(dp->write, dp->tab, dp->top, ep);
}

struct expr *
expr_mdifft(struct writer *write, size_t order, struct symtab *tab, 
		const struct sym **diffs, size_t diffsz, 
		const struct expr *ep)
{
	struct expr	*p;
	struct diff	 d;
	size_t		 i, j;

	p = expr_clone(ep);

	d.write = write;
	d.tab = tab;
	d.top = p;

	expr_write_open(write);

	for (i = 0; i < order; i++)
		for (j = 0; j < diffsz; j++) {
			d.sym = diffs[j];
			expr_reset(p);
			expr_constness(diffs[j], p);
			expr_write_start(write, diffs[j]);
			expr_reduce(write, tab, p, p);
			expr_diff(&d, p);
			expr_write_end(write, p);
		}

	expr_write_close(write);
	return(p);
}
