/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "extern.h"

struct expr *
expr_alloc(enum etype type)
{
	struct expr	*ep;

	ep = calloc(1, sizeof(struct expr));
	if (NULL == ep) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}

	ep->type = type;
	return(ep);
}

/*
 * Clear the local resources of an expression (its parameters: maxima,
 * minima, etc.) and zero it.
 */
void
expr_zero(struct expr *ep)
{
	size_t		 i;

	if (NULL == ep)
		return;

	if (ET_UNARY == ep->type) {
		for (i = 0; i < ep->val.un.parmsz; i++)
			expr_free(ep->val.un.parms[i]);
		free(ep->val.un.parms);
	}
	memset(ep, 0, sizeof(struct expr));
	ep->type = ET__MAX;
}

void
expr_free(struct expr *ep)
{
	size_t		 i;

	if (NULL == ep)
		return;

	switch (ep->type) {
	case (ET_UNARY):
		for (i = 0; i < ep->val.un.parmsz; i++)
			expr_free(ep->val.un.parms[i]);
		free(ep->val.un.parms);
		expr_free(ep->val.un.val);
		break;
	case (ET_BINARY):
		expr_free(ep->val.bin.lhs);
		expr_free(ep->val.bin.rhs);
		break;
	case (ET_VALUE):
		break;
	default:
		abort();
		/* NOTREACHED */
	}

	free(ep);
}

struct expr *
expr_clone(const struct expr *ep)
{
	struct expr	*p;
	size_t		 i;

	assert(ep);
	p = expr_alloc(ep->type);
	p->flags = ep->flags;
	/* But we're not been differentiated... */
	p->flags &= ~EXPR_DIFF;

	switch (p->type) {
	case (ET_VALUE):
		p->val.val.type = ep->val.val.type;
		if (VT_NUM == p->val.val.type)
			p->val.val.val.num = 
				ep->val.val.val.num;
		else
			p->val.val.val.sym = 
				ep->val.val.val.sym;
		break;
	case (ET_UNARY):
		p->val.un.type = ep->val.un.type;
		p->val.un.val = expr_clone(ep->val.un.val);
		p->val.un.parmsz = ep->val.un.parmsz;
		if (0 == p->val.un.parmsz)
			break;
		p->val.un.parms = calloc
			(p->val.un.parmsz, sizeof(struct expr *));
		for (i = 0; i < p->val.un.parmsz; i++) {
			assert(ep->val.un.parms[i]);
			p->val.un.parms[i] = expr_clone
				(ep->val.un.parms[i]);
		}
		break;
	case (ET_BINARY):
		p->val.bin.type = ep->val.bin.type;
		p->val.bin.lhs = expr_clone(ep->val.bin.lhs);
		p->val.bin.rhs = expr_clone(ep->val.bin.rhs);
		break;
	default:
		abort();
		/* NOTREACHED */
	}

	return(p);
}
