/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef	EXTERN_H
#define	EXTERN_H

#define	DIFFT_VER	"0.1.12"

enum	greek {
	GREEK_alpha,
	GREEK_theta,
	GREEK_omicron,
	GREEK_tau,
	GREEK_beta,
	GREEK_pi,
	GREEK_upsilon,
	GREEK_gamma,
	GREEK_phi,
	GREEK_delta,
	GREEK_kappa,
	GREEK_rho,
	GREEK_epsilon,
	GREEK_lambda,
	GREEK_chi,
	GREEK_mu,
	GREEK_sigma,
	GREEK_psi,
	GREEK_zeta,
	GREEK_nu,
	GREEK_omega,
	GREEK_eta,
	GREEK_xi,
	GREEK_Gamma,
	GREEK_Delta,
	GREEK_Theta,
	GREEK_Pi,
	GREEK_Phi,
	GREEK_Lambda,
	GREEK_Sigma,
	GREEK_Psi,
	GREEK_Xi,
	GREEK_Upsilon,
	GREEK_Omega,
	GREEK__MAX,
};

enum	rule {
	/* start of differentiation rules */
	RULE_PROD = 0, /* product rule */
	RULE_RECIP, /* reciprocal rule */
	RULE_QUOT, /* quotient rule */
	RULE_POWER, /* general power rule */
	RULE_LN, /* log rule */
	RULE_CONST, /* constant rule */
	RULE_COEFF, /* linear coeff rule */
	RULE_SUM, /* sum rule */
	RULE_TRIG, /* trigometric rules */
	RULE_CALC, /* fund. thm. calculus */
	RULE_ERF, /* error function rule */

	/* start of reduction rules */
	RULE_REDUCE,

	RULE_ABSR, /* absorb */
	RULE_IDENT, /* identity */
	RULE_PLN, /* property of ln */
	RULE_PEXP, /* property of exponentiation */
	RULE_ARITH, /* arithmetic */
	RULE__MAX
};

enum	etype {
	ET_VALUE = 0, /* constant or variable */
	ET_UNARY, /* a unary relation */
	ET_BINARY, /* a binary relation */
	ET__MAX
};

enum	btype {
	BT_ADD = 0, /* addition */
	BT_SUB, /* subtraction */
	BT_MUL, /* multiplication */
	BT_DIV, /* division */
	BT_EXP, /* exponentiation */
	BT__MAX
};

enum	utype {
	UT_NEG, /* negation */
	UT_LN, /* natural log */
	UT_SIN, /* sin */
	UT_COS, /* cosine */
	UT_TAN, /* tangent */
	UT_ARCCOS, /* inverse cosin */
	UT_ARCSIN, /* inverse sin */
	UT_ARCTAN, /* inverse tangent */
	UT_INTEGRAL, /* integral */
	UT_ERF, /* error function */
	UT__MAX
};

enum	vtype {
	VT_NUM, /* number */
	VT_IDENT, /* identifier */
	VT__MAX
};

struct	sym {
	char		*name; /* symbol name */
	struct sym	*next; /* next in list of symbols */
	enum greek	 greek; /* greek letter (or GREEK__MAX) */
};

struct	expr_unary {
	enum utype	  type;
	struct expr	 *val;
	struct expr	**parms;
	size_t		  parmsz;
};

struct	expr_binary {
	enum btype	 type;
	struct expr	*lhs;
	struct expr	*rhs;
};

struct	expr_value {
	enum vtype	 type;
	union {
		long double num; /* numerical */
		struct sym *sym; /* symbol */
	} val;
};

struct	expr {
	enum etype	 type;
	union {
		struct expr_unary un;
		struct expr_binary bin;
		struct expr_value val;
	} val;
	int		 flags;
#define	EXPR_CONST	(1 << 0) /* expr is constant */
#define	EXPR_DIFF	(1 << 1) /* expr being differentiated */
#define	EXPR_RDC	(1 << 2) /* expr is being reduced */
};

struct	symtab {
	struct sym	*syms;
};

enum	writert {
	WRITE_LATEX,
	WRITE_MATHML,
	WRITE_GNUPLOT,
	WRITE_C
};

struct	writer {
	void	(*write)(enum rule, const struct expr *, void *);
	void	(*write_close)(void *);
	void	(*write_end)(const struct expr *, void *);
	void	(*write_open)(void *);
	void	(*write_start)(const struct sym *, void *);
	void	(*write_free)(void *);
	void	 *arg;
};

__BEGIN_DECLS

struct expr	 *expr_alloc(enum etype);
struct expr	 *expr_clone(const struct expr *);
void		  expr_free(struct expr *);
struct expr	 *expr_parse(struct symtab *, 
			const char *, const char *);
void		  expr_reduce(struct writer *, struct symtab *, 
			const struct expr *, struct expr *);
struct expr	 *expr_mdifft(struct writer *, size_t, struct symtab *, 
			const struct sym **, size_t, const struct expr *);
struct writer	 *expr_write_c(char *, const char *,
			const struct sym **, size_t);
struct writer	 *expr_write_mathml(char *, int, size_t,
			const struct sym **, size_t,
			const struct sym **, size_t);
struct writer	 *expr_write_latex(char *, const char *, int, size_t,
			const struct sym **, size_t,
			const struct sym **, size_t);
struct writer	 *expr_write_gnuplot(char *, const char *,
			const struct sym **, size_t);
void		  expr_write_free(struct writer *);
void	 	  expr_write(struct writer *, enum rule, const struct expr *);
void	 	  expr_write_close(struct writer *);
void	 	  expr_write_end(struct writer *, const struct expr *);
void	 	  expr_write_open(struct writer *);
void	 	  expr_write_start(struct writer *, const struct sym *);
void		  expr_zero(struct expr *);
void		  symtab_free(struct symtab *);
struct sym	 *symtab_add(struct symtab *, const char *, size_t);
int		  symtab_has(struct symtab *, const char *, size_t);

__END_DECLS

#endif	/* ! EXTERN_H */
