/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "extern.h"

struct	latex {
	int		 	  verbose;
	size_t		 	  order;
	const struct expr	 *last;
	size_t			  domainsz;
	const struct sym	**domains;
	size_t			  diffsz;
	const struct sym	**diffs;
	const char		 *func;
	int			  flags;
#define	LATEX_FRAGMENT	 	(0x01)
};

static	void	 expr(const struct expr *, int);
static	void	 expr_value(const struct expr_value *, int);
static	void	 expr_binary(const struct expr_binary *, int);
static	void	 expr_unary(const struct expr_unary *, int);

static	void	 expr_latex_close(void *);
static	void	 expr_latex_end(const struct expr *, void *);
static	void	 expr_latex_free(void *);
static	void	 expr_latex_start(const struct sym *, void *);
static	void	 expr_latex_open(void *);
static	void	 expr_latex_write(enum rule, const struct expr *, void *);

static	const char * const rules[RULE__MAX] = {
	"D.PROD", /* RULE_PROD */
	"D.RCPR", /* RULE_RECIP */
	"D.QUOT", /* RULE_QUOT */
	"D.PWR", /* RULE_POWER */
	"D.LN", /* RULE_LN */
	"D.CONST", /* RULE_CONST */
	"D.COEF", /* RULE_COEFF */
	"D.SUM", /* RULE_SUM */
	"D.TRIG", /* RULE_TRIG */
	"D.CALC", /* RULE_CALC */
	"D.ERF", /* RULE_ERF */
	"",
	"R.ABSR", /* RULE_ABSR */
	"R.IDENT", /* RULE_IDENTITY */
	"R.LN", /* RULE_PLN */
	"R.EXP", /* RULE_PEXP */
	"R.ARITH", /* RULE_ARITH */
};

static void
expr_unary(const struct expr_unary *ep, int paren)
{

	if (paren)
		printf("\\left(");

	switch (ep->type) {
	case (UT_NEG):
		putchar('-');
		break;
	case (UT_LN):
		printf("\\ln");
		break;
	case (UT_ARCCOS):
		printf("\\arccos");
		break;
	case (UT_ARCSIN):
		printf("\\arcsin");
		break;
	case (UT_ARCTAN):
		printf("\\arctan");
		break;
	case (UT_SIN):
		printf("\\sin");
		break;
	case (UT_COS):
		printf("\\cos");
		break;
	case (UT_TAN):
		printf("\\tan");
		break;
	case (UT_ERF):
		printf("\\mathrm{erf}");
		break;
	case (UT_INTEGRAL):
		assert(3 == ep->parmsz);
		printf("\\int_");
		expr(ep->parms[0], 1);
		putchar('^');
		expr(ep->parms[1], 1);
		printf("\\!");
		expr(ep->val, 1);
		printf("\\,\\mathrm{d}");
		expr(ep->parms[2], 1);
		return;
	default:
		abort();
		/* NOTEACHED */
	}

	expr(ep->val, 1);
	if (paren)
		printf("\\right)");
}

static void
expr_binary(const struct expr_binary *ep, int paren)
{
	const char	*cp;

	cp = NULL;
	switch (ep->type) {
	case (BT_ADD):
		cp = "+";
		break;
	case (BT_SUB):
		cp = "-";
		break;
	case (BT_MUL):
		cp = "\\cdot";
		break;
	case (BT_EXP):
		cp = "^";
		paren = 0;
		if (ET_VALUE != ep->rhs->type)
			break;
		if (VT_NUM != ep->rhs->val.val.type)
			break;
		if (0.5L != ep->rhs->val.val.val.num)
			break;
		printf("\\sqrt");
		expr(ep->lhs, 0);
		return;
	case (BT_DIV):
		printf("\\frac");
		expr(ep->lhs, 0);
		expr(ep->rhs, 0);
		return;
	default:
		abort();
		/* NOTEACHED */
	}

	if (paren)
		printf("\\left(");

	expr(ep->lhs, 1);
	printf("%s", cp);
	expr(ep->rhs, 1);
	
	if (paren)
		printf("\\right)");
}

/* ARGSUSED */
static void
expr_value(const struct expr_value *ep, int paren)
{

	switch (ep->type) {
	case (VT_NUM):
		printf("%Lg", ep->val.num);
		break;
	case (VT_IDENT):
		if (GREEK__MAX != ep->val.sym->greek) 
			putchar('\\');
		printf("%s", ep->val.sym->name);
		break;
	default:
		abort();
		/* NOTREACHED */
	}
}

static void
expr(const struct expr *ep, int paren)
{

	assert(ep);

	putchar('{');
	if (EXPR_DIFF & ep->flags)
		printf("\\color{red}\\left[");
	else if (EXPR_RDC & ep->flags)
		printf("\\color{blue} ");
	
	switch (ep->type) {
	case (ET_BINARY):
		expr_binary(&ep->val.bin, paren);
		break;
	case (ET_UNARY):
		expr_unary(&ep->val.un, paren);
		break;
	case (ET_VALUE):
		expr_value(&ep->val.val, paren);
		break;
	default:
		abort();
		/* NOTREACHED */
	}

	if (EXPR_DIFF & ep->flags)
		printf("\\right]^\\prime");
	putchar('}');
}

/* ARGSUSED */
static void
expr_latex_write(enum rule r, const struct expr *ep, void *arg)
{
	struct latex	*p;

	p = (struct latex *)arg;
	if (0 == p->verbose)
		return;

	printf("= &");
	expr(ep, 0);
	printf(" & \\text{\\tiny{%s}} \\\\ \n", rules[r]);
}

/* ARGSUSED */
static void
expr_latex_open(void *arg)
{
	struct latex	*p;

	p = (struct latex *)arg;

	if ( ! (LATEX_FRAGMENT & p->flags)) {
		puts("\\documentclass[a4paper]{article}");
		puts("\\usepackage{amsmath,amsthm,amssymb,nopageno,xcolor}");
		puts("\\allowdisplaybreaks");
		puts("\\begin{document}");
	}
	puts("\\begin{align*}");

}
/* ARGSUSED */
static void
expr_latex_start(const struct sym *sym, void *arg)
{
	struct latex	*p;
	size_t		 i;

	p = (struct latex *)arg;

	if (0 == p->verbose)
		return;

	if (p->last)
		puts("& \\\\");

	printf("%s", p->func);
	if (p->order) {
		printf("^\\prime_{");
		if (GREEK__MAX != sym->greek) 
			putchar('\\');
		printf("%s}", sym->name);
	}

	putchar('(');
	for (i = 0; i < p->domainsz; i++) {
		if (i)
			putchar(',');
		if (GREEK__MAX != p->domains[i]->greek) 
			putchar('\\');
		printf("%s", p->domains[i]->name);
	}
	putchar(')');
}

/* ARGSUSED */
static void
expr_latex_end(const struct expr *ep, void *arg)
{
	struct latex	*p;

	p = (struct latex *)arg;
	p->last = ep;

	if (0 == p->verbose)
		return;

	printf("= &");
	expr(ep, 0);
}

/* ARGSUSED */
static void
expr_latex_close(void *arg)
{
	struct latex	*p;
	size_t		 i;

	p = (struct latex *)arg;

	/*
	 * If we're in verbose mode, we've already printed the last
	 * expression and can safely close shop and exit.
	 */

	if (p->verbose) {
		putchar('\n');
		puts("\\end{align*}");
		if ( ! (LATEX_FRAGMENT & p->flags))
			puts("\\end{document}");
		return;
	}

	/*
	 * Not in verbose mode: print out the last expression to be
	 * invoked.  This will be our "final" product.
	 */

	printf("%s", p->func);
	if (p->order) {
		printf("^%zu_{", p->order);
		for (i = 0; i < p->diffsz; i++) {
			if (i)
				putchar(' ');
			if (GREEK__MAX != p->diffs[i]->greek) 
				putchar('\\');
			printf("%s", p->diffs[i]->name);
		}
		putchar('}');
	}
	putchar('(');
	for (i = 0; i < p->domainsz; i++) {
		if (i)
			putchar(',');
		if (GREEK__MAX != p->domains[i]->greek) 
			putchar('\\');
		printf("%s", p->domains[i]->name);
	}
	printf(") = ");

	assert(p->last);
	expr(p->last, 0);

	putchar('\n');

	puts("\\end{align*}");

	if ( ! (LATEX_FRAGMENT & p->flags))
		puts("\\end{document}");
}

static void
expr_latex_free(void *arg)
{

	free(arg);
}

struct writer *
expr_write_latex(char *cp, const char *func,
		int verbose, size_t order,
		const struct sym **domains, size_t domainsz,
		const struct sym **diffs, size_t diffsz)
{
	struct writer	*p;
	struct latex	*arg;
	int		 flags;

	flags = 0;

	if (cp && 0 == strcmp(cp, "fragment"))
		flags |= LATEX_FRAGMENT;
	else if (cp) {
		fprintf(stderr, "%s: Invalid argument\n", cp);
		return(NULL);
	}

	arg = calloc(1, sizeof(struct latex));
	if (NULL == arg) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}
	arg->verbose = verbose;
	arg->order = order;
	arg->diffs = diffs;
	arg->diffsz = diffsz;
	arg->domains = domains;
	arg->domainsz = domainsz;
	arg->func = func;
	arg->flags = flags;

	p = calloc(1, sizeof(struct writer));
	if (NULL == p) {
		perror(NULL);
		free(arg);
		exit(EXIT_FAILURE);
	}

	p->write = expr_latex_write;
	p->write_close = expr_latex_close;
	p->write_end = expr_latex_end;
	p->write_open = expr_latex_open;
	p->write_start = expr_latex_start;
	p->write_free = expr_latex_free;
	p->arg = arg;

	return(p);
}
