/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "extern.h"

static	void	 usage(void);

int
main(int argc, char *argv[])
{
	char		 *buf, *bufp, *oargs;
	const char	 *xarg, *darg, *cp;
	const char	 *filen,  /* name of input file */
	      		 *function; /* name of function */
	size_t		  i, off, sz;
	int		  fd, ch;
	int		  nodiff, /* don't differentiate */
			  verb; /* verbose output */
	size_t		  order, /* order of differentiation */
			  diffsz, /* number of diff symbols */
			  domainsz; /* number of domain symbols */
	ssize_t		  ssz;
	struct expr	 *ep, *res;
	struct symtab	  tab; /* symbol table */
	struct sym	**diffs, /* differentiation symbols */
			**domains; /* domain symbols */
	struct writer	 *write;
	enum writert	  wt;
	extern int	  optind;
	extern char	 *optarg;

	memset(&tab, 0, sizeof(struct symtab));

	function = "f";
	fd = -1;
	nodiff = verb = 0;
	order = 1;
	oargs = buf = NULL;
	xarg = darg = NULL;
	wt = WRITE_LATEX;
	ep = res = NULL;
	write = NULL;
	sz = 1024;
	diffs = domains = NULL;

	while (-1 != (ch = getopt(argc, argv, "d:f:no:O:T:vx:")))
		switch (ch) {
		case ('d'):
			darg = optarg;
			break;
		case ('f'):
			function = optarg;
			break;
		case ('n'):
			nodiff = 1;
			break;
		case ('o'):
			if ((ch = atoi(optarg)) > 0) {
				order = (size_t)ch;
				break;
			}
			fprintf(stderr, "-o: Bad argument\n");
			goto out;
			/* NOTREACHED */
		case ('O'):
			oargs = optarg;
			break;
		case ('T'):
			if (0 == strcmp(optarg, "latex")) {
				wt = WRITE_LATEX;
				break;
			} else if (0 == strcmp(optarg, "mathml")) {
				wt = WRITE_MATHML;
				break;
			} else if (0 == strcmp(optarg, "gnuplot")) {
				wt = WRITE_GNUPLOT;
				break;
			} else if (0 == strcmp(optarg, "c")) {
				wt = WRITE_C;
				break;
			}
			fprintf(stderr, "-T: Bad argument\n");
			goto out;
			/* NOTREACHED */
		case ('v'):
			verb++;
			break;
		case ('x'):
			xarg = optarg;
			break;
		default:
			usage();
			/* NOTREACHED */
		}

	argc -= optind;
	argv += optind;

	/* Set default domain & differentiation symbols. */

	/* FIXME */

	if (NULL == xarg)
		xarg = "x";
	if (NULL == darg)
		darg = xarg;

	/* Without differentiating, order/diffs are useless. */

	if (nodiff && order) {
		fprintf(stderr, "-o: Ignored with -n\n");
		order = 0;
	} else if (nodiff && xarg) {
		fprintf(stderr, "-x: Ignored with -n\n");
		xarg = NULL;
	}

	/* Configure domain symbols. */

	assert(darg);
	domainsz = 1;
	for (cp = darg; (cp = strchr(cp, ',')); cp++)
		domainsz++;

	domains = malloc(domainsz * sizeof(struct sym *));
	i = 0;
	while (NULL != (cp = strchr(darg, ','))) {
		domains[i++] = symtab_add(&tab, darg, cp - darg);
		darg = cp + 1;
	}
	domains[i] = symtab_add(&tab, darg, strlen(darg));

	/* Configure differentiable symbols. */

	diffsz = 0;

	if (NULL != xarg) {
		diffsz = 1;
		for (cp = xarg; (cp = strchr(cp, ',')); cp++)
			diffsz++;

		diffs = malloc(diffsz * sizeof(struct sym *));
		i = 0;
		while (NULL != (cp = strchr(xarg, ','))) {
			diffs[i++] = symtab_add(&tab, xarg, cp - xarg);
			xarg = cp + 1;
		}

		diffs[i] = symtab_add(&tab, xarg, strlen(xarg));
	} 

	/* Sanitise our differentiation symbols. */

	for (i = 0; i < diffsz; i++) {
		if (strcmp("e", diffs[i]->name))
			continue;
		fprintf(stderr, "-x: Cannot use Euler's number\n");
		goto out;
	}

	/* Configure output writers. */

	if (WRITE_LATEX == wt)
		write = expr_write_latex
			(oargs, function, verb, order, 
			 (const struct sym **)domains, domainsz,
			 (const struct sym **)diffs, diffsz);
	else if (WRITE_MATHML == wt)
		write = expr_write_mathml
			(oargs, verb, order,
			 (const struct sym **)domains, domainsz,
			 (const struct sym **)diffs, diffsz);
	else if (WRITE_GNUPLOT == wt)
		write = expr_write_gnuplot
			(oargs, function,
			 (const struct sym **)domains, domainsz);
	else 
		write = expr_write_c
			(oargs, function,
			 (const struct sym **)domains, domainsz);

	if (NULL == write)
		goto out;

	/* Now read our input into one character buffer. */

	fd = STDIN_FILENO;
	filen = "<stdin>";

	if (argc) {
		filen = *argv++;
		fd = open(filen, O_RDONLY, 0);
		if (-1 == fd) {
			perror(filen);
			goto out;
		}
	}

	if (NULL == (buf = malloc(sz))) {
		perror(NULL);
		goto out;
	}

	for (off = 0; ; ) {
		if (off == sz) {
			sz += 1024;
			if (NULL == (bufp = realloc(buf, sz))) {
				perror(NULL);
				goto out;
			}
			buf = bufp;
		}

		ssz = read(fd, buf + (int)off, sz - off);
		if (ssz == 0) {
			sz = off;
			break;
		} else if (ssz == -1) {
			perror(filen);
			goto out;
		}
		off += (size_t)ssz;
	}

	buf[(int)off] = '\0';

	/* Parse the input buffer into an expression. */

	if (NULL == (ep = expr_parse(&tab, filen, buf)))
		goto out;

	/* Either write it out or differentiate and write out. */

	if (nodiff) {
		expr_write_open(write);
		expr_write_start(write, diffs[0]);
		expr_reduce(write, &tab, ep, ep);
		expr_write_end(write, ep);
		expr_write_close(write);
	} else
		res = expr_mdifft
			(write, order, &tab, 
			 (const struct sym **)diffs, diffsz, ep);
out:
	/* Run all closure handlers. */

	if (-1 != fd && STDIN_FILENO != fd)
		close(fd);
	if (buf)
		free(buf);
	if (ep)
		expr_free(ep);
	if (res)
		expr_free(res);
	if (write)
		expr_write_free(write);
	if (diffs)
		free(diffs);
	if (domains)
		free(domains);

	symtab_free(&tab);
	return(ep ? EXIT_SUCCESS : EXIT_FAILURE);
}

static void
usage(void)
{
	extern char	*__progname;

	fprintf(stderr, "usage: %s [-nv] "
				  "[-d syms] "
				  "[-f function] "
				  "[-o order] "
				  "[-O options] "
				  "[-T mode] "
				  "[-x syms] "
				  "[file]\n", __progname);
	exit(EXIT_FAILURE);
}
