/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#if defined(__linux__) || defined(__MINT__)
# define _GNU_SOURCE /* strptime(), getsubopt() */
#endif

#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "extern.h"

#define	UNCONST(a)	((void *)(uintptr_t)(const void *)(a))

struct	mathml {
	const struct expr	 *last;
	int			  verbose;
	int	 		  start;
	size_t	 		  order;
	const char		 *style;
	size_t			  diffsz;
	const struct sym	**diffs;
	size_t			  domainsz;
	const struct sym	**domains;
};

static	void	 expr(const struct expr *, size_t);
static	void	 expr_value(const struct expr_value *, size_t);
static	void	 expr_binary(const struct expr_binary *, size_t);
static	void	 expr_unary(const struct expr_unary *, size_t);

static	void	 expr_tabs(size_t);

static	void	 expr_mathml_close(void *);
static	void	 expr_mathml_end(const struct expr *, void *);
static	void	 expr_mathml_free(void *);
static	void	 expr_mathml_start(const struct sym *, void *);
static	void	 expr_mathml_open(void *);
static	int	 expr_mathml_options(struct mathml *, char *);
static	void	 expr_mathml_write(enum rule, const struct expr *, void *);

static	const char *const rules[RULE__MAX] = {
	"product", /* RULE_PROD */
	"reciprocal", /* RULE_RECIP */
	"quotient", /* RULE_QUOT */
	"power", /* RULE_POWER */
	"logarithm", /* RULE_LN */
	"constant", /* RULE_CONST */
	"coefficient", /* RULE_COEFF */
	"sum", /* RULE_SUM */
	"trig", /* RULE_TRIG */
	"calc", /* RULE_CALC */
	"erf", /* RULE_ERF */
	"",
	"absorption", /* RULE_ABSR */
	"identity", /* RULE_IDENTITY */
	"logarithm", /* RULE_PLN */
	"exponet", /* RULE_PEXP */
	"arithmetic", /* RULE_ARITH */
};

static	const char * const unaries[UT__MAX] = {
	"-", /* UT_NEG */
	"ln", /* UT_LN */
	"sin", /* UT_SIN */
	"cos", /* UT_COS */
	"tan", /* UT_TAN */
	"arccos", /* UT_ARCCOS */
	"arcsin", /* UT_ARCSIN */
	"arctan", /* UT_ARCTAN */
	NULL, /* UT_INTEGRAL */
	"erf" /* UT_ERF */
};

static void
expr_tabs(size_t tabs)
{
	size_t	 	 i, j;

	for (i = 0; i < tabs; i++)
		for (j = 0; j < 4; j++)
			putchar(' ');
}

static void
expr_unary(const struct expr_unary *ep, size_t tabs)
{

	if (UT_INTEGRAL == ep->type) {
		expr_tabs(tabs++);
		puts("<msubsup>");
		expr_tabs(tabs);
		puts("<mo>&int;</mo>");
		expr(ep->parms[0], tabs);
		expr(ep->parms[1], tabs);
		expr_tabs(--tabs);
		puts("</msubsup>");
		expr(ep->val, tabs);
		expr_tabs(tabs++);
		puts("<mrow>");
		expr_tabs(tabs);
		puts("<mo>&dd;</mo>");
		expr(ep->parms[2], tabs);
		expr_tabs(--tabs);
		puts("</mrow>");
		return;
	}

	assert(ep->type < UT__MAX);
	expr_tabs(tabs);
	printf("<mo>%s</mo>\n", unaries[ep->type]);
	expr(ep->val, tabs);
}

static void
expr_binary(const struct expr_binary *ep, size_t tabs)
{
	const char	*cp;

	cp = NULL;
	switch (ep->type) {
	case (BT_ADD):
		cp = "+";
		break;
	case (BT_SUB):
		cp = "-";
		break;
	case (BT_MUL):
		cp = "&middot;";
		break;
	case (BT_DIV):
		expr_tabs(tabs);
		puts("<mfrac>");
		expr(ep->lhs, tabs + 1);
		expr(ep->rhs, tabs + 1);
		expr_tabs(tabs);
		puts("</mfrac>");
		return;
	case (BT_EXP):
		if (ET_VALUE == ep->rhs->type)
			if (VT_NUM == ep->rhs->val.val.type)
				if (0.5L == ep->rhs->val.val.val.num) {
					expr_tabs(tabs);
					puts("<msqrt>");
					expr(ep->lhs, tabs + 1);
					expr_tabs(tabs);
					puts("</msqrt>");
					return;
				}
		expr_tabs(tabs);
		puts("<msup>");
		expr(ep->lhs, tabs + 1);
		expr(ep->rhs, tabs + 1);
		expr_tabs(tabs);
		puts("</msup>");
		return;
	default:
		abort();
		/* NOTEACHED */
	}

	expr_tabs(tabs++);
	puts("<mrow>");

	expr_tabs(tabs);
	puts("<mo>(</mo>");

	expr(ep->lhs, tabs);

	expr_tabs(tabs);
	printf("<mo>%s</mo>\n", cp);

	expr(ep->rhs, tabs);

	expr_tabs(tabs);
	puts("<mo>)</mo>");

	expr_tabs(--tabs);
	puts("</mrow>");
}

static void
expr_value(const struct expr_value *ep, size_t tabs)
{

	expr_tabs(tabs);

	switch (ep->type) {
	case (VT_NUM):
		printf("<mn>%Lg</mn>\n", ep->val.num);
		break;
	case (VT_IDENT):
		printf("<mi>");
		if (GREEK__MAX != ep->val.sym->greek) 
			putchar('&');
		printf("%s", ep->val.sym->name);
		if (GREEK__MAX != ep->val.sym->greek) 
			putchar(';');
		printf("</mi>\n");
		break;
	default:
		abort();
		/* NOTREACHED */
	}
}

static void
expr(const struct expr *ep, size_t tabs)
{

	expr_tabs(tabs++);
	
	if (EXPR_DIFF & ep->flags)
		puts("<mrow class=\"diff\">");
	else if (EXPR_RDC & ep->flags)
		puts("<mrow class=\"reduce\">");
	else
		puts("<mrow>");

	switch (ep->type) {
	case (ET_BINARY):
		expr_binary(&ep->val.bin, tabs);
		break;
	case (ET_UNARY):
		expr_unary(&ep->val.un, tabs);
		break;
	case (ET_VALUE):
		expr_value(&ep->val.val, tabs);
		break;
	default:
		abort();
		/* NOTREACHED */
	}
	expr_tabs(--tabs);
	puts("</mrow>");
}

/* ARGSUSED */
static void
expr_mathml_write(enum rule r, const struct expr *ep, void *arg)
{
	size_t		 i;
	struct mathml	*m;

	m = (struct mathml *)arg;

	if (0 == m->verbose)
		return;

	if (0 == m->start) {
		i = 0;
		expr_tabs(i++);
		puts("<mtr>");
		expr_tabs(i++);
		puts("<mtd></mtd>");
	} else
		i = 2;

	m->start = 0;

	expr_tabs(i);
	puts("<mtd><mo>=</mo></mtd>");
	expr_tabs(i++);
	puts("<mtd columnalign=\"left\">");

	expr(ep, i);

	expr_tabs(--i);
	puts("</mtd>");
	expr_tabs(i++);
	puts("<mtd columnalign=\"right\">");

	if (r < RULE__MAX) {
		expr_tabs(i);
		printf("<mtext class=\"%s\" "
				"mathsize=\"small\">%s</mtext>\n", 
				r < RULE_REDUCE ? "diff" : "reduce",
				rules[r]);
	}

	expr_tabs(--i);
	puts("</mtd>");
	expr_tabs(--i);
	puts("</mtr>");
}

/* ARGSUSED */
static void
expr_mathml_open(void *arg)
{
	struct mathml	*m;

	m = (struct mathml *)arg;

	puts("<?xml version=\"1.0\" "
			"encoding=\"UTF-8\"?>");
	puts("<!DOCTYPE html PUBLIC "
			"\"-//W3C//DTD XHTML 1.1 plus MathML 2.0//EN\" "
			"\"http://www.w3.org/Math/DTD/mathml2/xhtml-math11-f.dtd\">");
	puts("<html xmlns=\"http://www.w3.org/1999/xhtml\">");
	puts("<head>");

	if (m->style) {
		expr_tabs(1);
		printf("<link rel=\"stylesheet\" "
				"href=\"%s\" "
				"type=\"text/css\" "
				"media=\"all\" />\n", m->style);
	}

	puts("</head>");
	puts("<body>");
	puts("<math xmlns=\"http://www.w3.org/1998/Math/MathML\" "
			"display=\"block\">");
	puts("<mtable>");
}

/* ARGSUSED */
static void
expr_mathml_start(const struct sym *sym, void *arg)
{
	struct mathml	*m;
	size_t		 i, j;

	m = (struct mathml *)arg;
	if (0 == m->verbose)
		return;

	m->start = 1;
	i = 0;

	expr_tabs(i++);
	puts("<mtr>");
	expr_tabs(i++);
	puts("<mtd>");
	expr_tabs(i++);
	puts("<mrow>");
	expr_tabs(i);

	if (m->order) {
		puts("<msubsup>");
		expr_tabs(++i);
		puts("<mi>f</mi>");
		expr_tabs(i);
		printf("<mi>");
		if (GREEK__MAX != sym->greek) 
			putchar('&');
		printf("%s", sym->name);
		if (GREEK__MAX != sym->greek) 
			putchar(';');
		puts("</mi>");
		expr_tabs(i);
		puts("<mn>&prime;</mn>");
		expr_tabs(--i);
		puts("</msubsup>");
	} else
		puts("<mi>f</mi>");

	expr_tabs(i);
	puts("<mo>&#x2061;</mo>");
	expr_tabs(i++);
	puts("<mrow>");
	expr_tabs(i);
	puts("<mo>(</mo>");
	for (j = 0; j < m->domainsz; j++) {
		expr_tabs(i);
		printf("<mi>");
		if (GREEK__MAX != m->domains[j]->greek) 
			putchar('&');
		printf("%s", m->domains[j]->name);
		if (GREEK__MAX != m->domains[j]->greek) 
			putchar(';');
		puts("</mi>");
	}
	expr_tabs(i);
	puts("<mo>)</mo>");
	expr_tabs(--i);
	puts("</mrow>");
	expr_tabs(--i);
	puts("</mrow>");
	expr_tabs(--i);
	puts("</mtd>");
}

/* ARGSUSED */
static void
expr_mathml_end(const struct expr *ep, void *arg)
{
	struct mathml	*m;

	m = (struct mathml *)arg;
	m->last = ep;
	if (0 == m->verbose)
		return;

	expr_mathml_write(RULE__MAX, ep, arg);
	m->start = 0;
}

/* ARGSUSED */
static void
expr_mathml_close(void *arg)
{
	struct mathml	*m;
	size_t		 i, j;

	m = (struct mathml *)arg;

	if (m->verbose) {
		puts("</mtable>");
		puts("</math>");
		puts("</body>");
		puts("</html>");
		return;
	}

	assert(m->last);

	i = 0;
	expr_tabs(i++);
	puts("<mtr>");
	expr_tabs(i++);
	puts("<mtd>");
	expr_tabs(i++);
	puts("<mrow>");

	expr_tabs(i);
	if (m->order) {
		puts("<msubsup>");
		expr_tabs(++i);
		puts("<mi>f</mi>");
		expr_tabs(i++);
		puts("<mrow>");
		for (j = 0; j < m->diffsz; j++) {
			expr_tabs(i);
			printf("<mi>");
			if (GREEK__MAX != m->diffs[j]->greek) 
				putchar('&');
			printf("%s", m->diffs[j]->name);
			if (GREEK__MAX != m->diffs[j]->greek) 
				putchar(';');
			puts("</mi>");
		}
		expr_tabs(--i);
		puts("</mrow>");
		expr_tabs(i);
		printf("<mn>%zu</mn>\n", m->order);
		expr_tabs(--i);
		puts("</msubsup>");
	} else
		puts("<mi>f</mi>");

	expr_tabs(i);
	puts("<mo>&#x2061;</mo>");
	expr_tabs(i++);
	puts("<mrow>");
	expr_tabs(i);
	puts("<mo>(</mo>");

	for (j = 0; j < m->domainsz; j++) {
		expr_tabs(i);
		printf("<mi>");
		if (GREEK__MAX != m->domains[j]->greek) 
			putchar('&');
		printf("%s", m->domains[j]->name);
		if (GREEK__MAX != m->domains[j]->greek) 
			putchar(';');
		puts("</mi>");
	}

	expr_tabs(i);
	puts("<mo>)</mo>");
	
	expr_tabs(--i);
	puts("</mrow>");

	expr_tabs(i);
	puts("<mo>=</mo>");

	expr(m->last, i);

	expr_tabs(--i);
	puts("</mrow>");
	puts("</mtd>");
	puts("</mtr>");
	puts("</mtable>");
	puts("</math>");
	puts("</body>");
	puts("</html>");
}

static void
expr_mathml_free(void *arg)
{

	free(arg);
}

static int
expr_mathml_options(struct mathml *p, char *arg)
{
	char		*v, *o;
	const char	*toks[2]; 

	toks[0] = "style";
	toks[1] = NULL;

	if (NULL == arg)
		return(1);

	while (*arg) {
		o = arg;
		switch (getsubopt(&arg, UNCONST(toks), &v)) {
		case (0):
			p->style = v;
			break;
		default:
			fprintf(stderr, "-O%s: Bad argument\n", o);
			return(0);
		}
	}

	return(1);
}

/* ARGSUSED */
struct writer *
expr_write_mathml(char *cp, int verbose, size_t order,
		const struct sym **domains, size_t domainsz,
		const struct sym **diffs, size_t diffsz)
{
	struct writer	*p;
	struct mathml	*arg;

	arg = calloc(1, sizeof(struct mathml));
	if (NULL == arg) {
		perror(NULL);
		exit(EXIT_FAILURE);
	}
	arg->verbose = verbose;
	arg->order = order;
	arg->diffs = diffs;
	arg->diffsz = diffsz;
	arg->domains = domains;
	arg->domainsz = domainsz;

	p = calloc(1, sizeof(struct writer));
	if (NULL == p) {
		perror(NULL);
		free(arg);
		exit(EXIT_FAILURE);
	}

	p->write = expr_mathml_write;
	p->write_close = expr_mathml_close;
	p->write_end = expr_mathml_end;
	p->write_open = expr_mathml_open;
	p->write_start = expr_mathml_start;
	p->write_free = expr_mathml_free;
	p->arg = arg;

	if ( ! expr_mathml_options(arg, cp)) {
		free(arg);
		free(p);
		return(NULL);
	}

	return(p);
}
