/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "extern.h"

enum	eerr {
	ER_BADNUM = 0,
	ER_BADEXP,
	ER_BADVAL,
	ER_SUBEXP,
	ER__MAX
};

struct	parse {
	size_t		 ln;
	size_t		 er;
	size_t		 col;
	const char	*fn;
	const char	*dat;
	struct symtab	*sym;
};

static	const char *const unaries[UT__MAX] = {
	"-", /* UT_NEG */
	"ln", /* UT_LN */
	"sin", /* UT_SIN */
	"cos", /* UT_COS */
	"tan", /* UT_TAN */
	"arccos", /* UT_ARCCOS */
	"arcsin", /* UT_ARCSIN */
	"arctan", /* UT_ARCTAN */
	"integral", /* UT_INTEGRAL */
	"erf", /* UT_ERF */
};

static	const char *const binaries[BT__MAX] = {
	"+", /* BT_ADD */
	"-", /* BT_SUB */
	"*", /* BT_MUL */
	"/", /* BT_DIV */
	"^" /* BT_EXP */
};

static	const char *const eerrs[ER__MAX] = {
	"Malformed number", /* ER_BADNUM */
	"Bad expression syntax", /* ER_BADEXP */
	"Bad value syntax", /* ER_BADVAL */
	"Error in subexpression", /* ER_SUBEXP */
};

static	int	 	 cur(struct parse *);
static	struct expr	*expr_integral(struct parse *);
static	struct expr	*expr_number(struct parse *);
static	struct expr	*expr_string(struct parse *);
static	struct expr	*expr(struct parse *);
static	void	 	 fail(struct parse *, enum eerr);
static	int	 	 next(struct parse *);
static	int	 	 match(const char *, struct parse *);

static int
cur(struct parse *pp)
{

	return(*pp->dat);
}

static int
match(const char *p, struct parse *pp)
{
	size_t		 i;

	for (i = 0; *p; p++, i++)
		if (*p != pp->dat[i])
			return(0);
	pp->dat += i;
	pp->col += i;
	return(1);
}

static int
next(struct parse *pp)
{

	assert('\0' != *pp->dat);

	if ( ! isspace((int)*pp->dat)) {
		pp->dat++;
		pp->col++;
	} else
		while (isspace((int)*pp->dat)) {
			pp->dat++;
			pp->col++;
		}

	return((int)*pp->dat);
}

static void
fail(struct parse *pp, enum eerr er)
{

	if (pp->er++)
		return;
	fprintf(stderr, "%s:%zu:%zu: %s.\n",
		pp->fn, pp->ln + 1, pp->col + 1, eerrs[er]);
}

static struct expr *
expr_string(struct parse *pp)
{
	struct expr	*ep;
	const char	*sv;

	errno = 0;

	ep = expr_alloc(ET_VALUE);
	ep->val.val.type = VT_IDENT;

	sv = pp->dat;
	while (isalnum(cur(pp)))
		next(pp);

	ep->val.val.val.sym = symtab_add(pp->sym, sv, pp->dat - sv);
	return(ep);
}

static struct expr *
expr_number(struct parse *pp)
{
	char		*endp;
	struct expr	*ep;

	errno = 0;

	ep = expr_alloc(ET_VALUE);
	ep->val.val.type = VT_NUM;
	ep->val.val.val.num = strtold(pp->dat, &endp);

	if (endp == pp->dat || ERANGE == errno) {
		fail(pp, ER_BADNUM);
		expr_free(ep);
		return(NULL);
	}

	pp->col += endp - pp->dat;
	pp->dat = endp;
	return(ep);
}

static struct expr *
expr_integral(struct parse *pp)
{
	struct expr	*ep;

	ep = expr_alloc(ET_UNARY);

	ep->val.un.type = UT_INTEGRAL;
	ep->val.un.parmsz = 3;
	ep->val.un.parms = calloc
		(ep->val.un.parmsz, sizeof(struct expr *));

	if (NULL == (ep->val.un.parms[0] = expr(pp))) {
		expr_free(ep);
		return(NULL);
	}
	if (NULL == (ep->val.un.parms[1] = expr(pp))) {
		expr_free(ep);
		return(NULL);
	}
	if (NULL == (ep->val.un.val = expr(pp))) {
		expr_free(ep);
		return(NULL);
	}
	if (NULL == (ep->val.un.parms[2] = expr(pp))) {
		expr_free(ep);
		return(NULL);
	}
	if (ET_VALUE != ep->val.un.parms[2]->type) {
		fail(pp, ER_BADEXP);
		expr_free(ep);
		return(NULL);
	}
	if (VT_IDENT != ep->val.un.parms[2]->val.val.type) {
		fail(pp, ER_BADEXP);
		expr_free(ep);
		return(NULL);
	}
	return(ep);
}


static struct expr *
expr(struct parse *pp)
{
	struct expr	*ep, *lhs;
	size_t		 i;

	while (isspace(cur(pp)))
		next(pp);

	if ('\0' == cur(pp))
		return(NULL);

	for (i = 0; i < UT__MAX; i++)
		if (match(unaries[i], pp)) {
			if (UT_INTEGRAL == (enum utype)i)
				return(expr_integral(pp));
			ep = expr_alloc(ET_UNARY);
			ep->val.un.type = (enum utype)i;
			if (NULL != (ep->val.un.val = expr(pp)))
				return(ep);
			expr_free(ep);
			return(NULL);
		}

	if ('(' != cur(pp)) {
		if (isdigit(cur(pp)))
			return(expr_number(pp));
		else if (isalpha(cur(pp)))
			return(expr_string(pp));

		fail(pp, ER_BADEXP);
		return(NULL);
	} else
		next(pp);

	if (NULL == (lhs = expr(pp)))
		return(NULL);

	while (isspace(cur(pp)))
		next(pp);

	if (')' == cur(pp)) {
		next(pp);
		return(lhs);
	}

	ep = expr_alloc(ET_BINARY);
	ep->val.bin.lhs = lhs;

	for (i = 0; i < (size_t)BT__MAX; i++)
		if (match(binaries[i], pp))
			break;

	if (i == (size_t)BT__MAX) {
		fail(pp, ER_BADEXP);
		expr_free(ep);
		return(NULL);
	}

	ep->val.bin.type = (enum btype)i;
	ep->val.bin.rhs = expr(pp);

	if (NULL == ep->val.bin.rhs) {
		expr_free(ep);
		return(NULL);
	}

	while (isspace(cur(pp)))
		next(pp);

	if (')' != cur(pp)) {
		fail(pp, ER_BADEXP);
		expr_free(ep);
		return(NULL);
	}

	next(pp);
	return(ep);
}

struct expr *
expr_parse(struct symtab *sp, const char *fn, const char *cp)
{
	struct parse	 p;
	struct expr	*ep;

	memset(&p, 0, sizeof(struct parse));

	p.fn = fn;
	p.dat = cp;
	p.sym = sp;

	ep = expr(&p);
	while (isspace(cur(&p)))
		next(&p);

	if ('\0' != cur(&p)) {
		fail(&p, ER_BADEXP);
		expr_free(ep);
		ep = NULL;
	}

	return(ep);
}
