/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "extern.h"

struct	rdc {
	struct writer		*write;
	struct symtab		*tab;
	const struct expr	*top;
};

#define	WRITE(d, e, r) \
	do { \
		(e)->flags |= EXPR_RDC; \
		expr_write((d)->write, (r), (d)->top); \
		(e)->flags &= ~EXPR_RDC; \
	} while (0 /* CONSTCOND */)

static	int	 expr_rdc_addsub(const struct rdc *, struct expr *);
static	int	 expr_rdc_addsubr(const struct rdc *, 
			struct expr *, struct expr *);
static	int	 expr_rdc_bin(const struct rdc *, struct expr *);
static	int	 expr_rdc_div(const struct rdc *, struct expr *);
static	int	 expr_rdc_exp(const struct rdc *, struct expr *);
static	int	 expr_rdc_ign(const struct rdc *, struct expr *);
static	int	 expr_rdc_ln(const struct rdc *, struct expr *);
static	int	 expr_rdc_mul(const struct rdc *, struct expr *);
static	int	 expr_rdc_mulr(const struct rdc *, 
			struct expr *, struct expr *);
static	int	 expr_rdc_neg(const struct rdc *, struct expr *);
static	int	 expr_rdc_un(const struct rdc *, struct expr *);
static	void	 expr_rdc(const struct rdc *, struct expr *);

typedef	int	(*rdc)(const struct rdc *, struct expr *);

static	const	 rdc expr_rdcs[ET__MAX] = {
	expr_rdc_ign, /* ET_VALUE */
	expr_rdc_un, /* ET_UNARY */
	expr_rdc_bin /* ET_BINARY */
};

static	const	 rdc expr_un_rdcs[UT__MAX] = {
	expr_rdc_neg, /* UT_NEG */
	expr_rdc_ln, /* UT_LN */
	expr_rdc_ign, /* UT_SIN */
	expr_rdc_ign, /* UT_COS */
	expr_rdc_ign, /* UT_TAN */
	expr_rdc_ign, /* UT_ARCSIN */
	expr_rdc_ign, /* UT_ARCTAN */
	expr_rdc_ign, /* UT_INTEGRAL */
	expr_rdc_ign /* UT_ERF */
};

static int
expr_rdc_ign(const struct rdc *rp, struct expr *ep)
{

	return(0);
}

static int
expr_rdc_div(const struct rdc *rp, struct expr *ep)
{
	struct expr	*lhs, *rhs;

	lhs = ep->val.bin.lhs;
	rhs = ep->val.bin.rhs;

	switch (rhs->type) {
	case (ET_VALUE):
		switch (rhs->val.val.type) {
		case (VT_NUM):
			if (1.0L == rhs->val.val.val.num) {
				WRITE(rp, ep, RULE_IDENT);
				memcpy(ep, lhs, sizeof(struct expr));
				free(lhs);
				expr_free(rhs);
				return(1);
			} 
			break;
		default:
			break;
		}
	default:
		break;
	}

	if (ET_VALUE == lhs->type && ET_VALUE == rhs->type &&
			VT_NUM == lhs->val.val.type &&
			VT_NUM == rhs->val.val.type) {
		WRITE(rp, ep, RULE_ARITH);
		ep->type = ET_VALUE;
		ep->val.val.type = VT_NUM;
		ep->val.val.val.num =
			lhs->val.val.val.num / 
			rhs->val.val.val.num;
		expr_free(lhs);
		expr_free(rhs);
		return(1);
	}

	return(0);
}

static int
expr_rdc_exp(const struct rdc *rp, struct expr *ep)
{
	struct expr	*lhs, *rhs;

	lhs = ep->val.bin.lhs;
	rhs = ep->val.bin.rhs;

	switch (rhs->type) {
	case (ET_BINARY):
		switch (rhs->val.bin.type) {
		case (BT_ADD):
			WRITE(rp, ep, RULE_PEXP);
			ep->val.bin.type = BT_MUL;
			ep->val.bin.lhs = expr_alloc(ET_BINARY);
			ep->val.bin.lhs->val.bin.type = BT_EXP;
			ep->val.bin.lhs->val.bin.lhs = expr_clone(lhs);
			ep->val.bin.lhs->val.bin.rhs = rhs->val.bin.lhs;
			ep->val.bin.rhs = expr_alloc(ET_BINARY);
			ep->val.bin.rhs->val.bin.type = BT_EXP;
			ep->val.bin.rhs->val.bin.lhs = lhs;
			ep->val.bin.rhs->val.bin.rhs = rhs->val.bin.rhs;
			free(rhs);
			return(1);
		default:
			break;
		}
		break;
	case (ET_VALUE):
		switch (rhs->val.val.type) {
		case (VT_NUM):
			if (1.0L == rhs->val.val.val.num) {
				WRITE(rp, ep, RULE_PEXP);
				memcpy(ep, lhs, sizeof(struct expr));
				free(lhs);
				expr_free(rhs);
				return(1);
			} else if (0.0L == rhs->val.val.val.num) {
				WRITE(rp, ep, RULE_PEXP);
				ep->type = ET_VALUE;
				ep->val.val.type = VT_NUM;
				ep->val.val.val.num = 1.0L;
				expr_free(rhs);
				expr_free(lhs);
				return(1);
			}
			break;
		default:
			break;
		}
	default:
		break;
	}

	switch (lhs->type) {
	case (ET_BINARY):
		switch (lhs->val.bin.type) {
		case (BT_MUL):
			WRITE(rp, ep, RULE_PEXP);
			ep->val.bin.type = BT_MUL;
			ep->val.bin.lhs = expr_alloc(ET_BINARY);
			ep->val.bin.lhs->val.bin.type = BT_EXP;
			ep->val.bin.lhs->val.bin.lhs = lhs->val.bin.lhs;
			ep->val.bin.lhs->val.bin.rhs = rhs;
			ep->val.bin.rhs = expr_alloc(ET_BINARY);
			ep->val.bin.rhs->val.bin.type = BT_EXP;
			ep->val.bin.rhs->val.bin.lhs = lhs->val.bin.rhs;
			ep->val.bin.rhs->val.bin.rhs = expr_clone(rhs);
			free(lhs);
			return(1);
		case (BT_EXP):
			WRITE(rp, ep, RULE_PEXP);
			memcpy(ep, lhs, sizeof(struct expr));
			free(lhs);
			lhs = ep->val.bin.rhs;
			ep->val.bin.rhs = expr_alloc(ET_BINARY);
			ep->val.bin.rhs->val.bin.type = BT_MUL;
			ep->val.bin.rhs->val.bin.lhs = lhs;
			ep->val.bin.rhs->val.bin.rhs = rhs;
			return(1);
		default:
			break;
		}
		break;
	default:
		break;
	}

	if (ET_VALUE == lhs->type && ET_VALUE == rhs->type &&
			VT_NUM == lhs->val.val.type &&
			VT_NUM == rhs->val.val.type) {
		WRITE(rp, ep, RULE_ARITH);
		ep->type = ET_VALUE;
		ep->val.val.type = VT_NUM;
		ep->val.val.val.num =
			powl(lhs->val.val.val.num,
			     rhs->val.val.val.num);
		expr_free(lhs);
		expr_free(rhs);
		return(1);
	}

	return(0);
}

static int
expr_rdc_mul(const struct rdc *rp, struct expr *ep)
{
	struct expr	*lhs, *rhs;

	lhs = ep->val.bin.lhs;
	rhs = ep->val.bin.rhs;

	/* Absorbtion of multiplication. */

	if ((ET_VALUE == lhs->type && 
			 VT_NUM == lhs->val.val.type && 
			 0.0L == lhs->val.val.val.num) ||
			(ET_VALUE == rhs->type && 
			 VT_NUM == rhs->val.val.type && 
			 0.0L == rhs->val.val.val.num)) {
		WRITE(rp, ep, RULE_ABSR);
		expr_free(lhs);
		expr_free(rhs);
		ep->type = ET_VALUE;
		ep->val.val.type = VT_NUM;
		ep->val.val.val.num = 0.0L;
		return(1);
	}

	/* Identity of multiplication (lhs). */

	if (ET_VALUE == lhs->type && 
			VT_NUM == lhs->val.val.type && 
			1.0L == lhs->val.val.val.num) {
		WRITE(rp, ep, RULE_IDENT);
		expr_free(lhs);
		memcpy(ep, rhs, sizeof(struct expr));
		free(rhs);
		return(1);
	}

	/* Identity of multiplication (rhs). */

	if (ET_VALUE == rhs->type && 
			VT_NUM == rhs->val.val.type && 
			1.0L == rhs->val.val.val.num) {
		WRITE(rp, ep, RULE_IDENT);
		expr_free(rhs);
		memcpy(ep, lhs, sizeof(struct expr));
		free(lhs);
		return(1);
	}

	/* Negative identity of multiplication (lhs). */

	if (ET_VALUE == lhs->type && 
			VT_NUM == lhs->val.val.type && 
			-1.0L == lhs->val.val.val.num) {
		WRITE(rp, ep, RULE_IDENT);
		expr_free(lhs);
		expr_zero(ep);
		ep->type = ET_UNARY;
		ep->val.un.type = UT_NEG;
		ep->val.un.val = rhs;
		return(1);
	}

	/* Negative identity of multiplication (rhs). */

	if (ET_VALUE == rhs->type && 
			VT_NUM == rhs->val.val.type && 
			-1.0L == rhs->val.val.val.num) {
		WRITE(rp, ep, RULE_IDENT);
		expr_free(rhs);
		expr_zero(ep);
		ep->type = ET_UNARY;
		ep->val.un.type = UT_NEG;
		ep->val.un.val = lhs;
		return(1);
	}

	/* Multiplicative arithmetic. */

	if (ET_VALUE == rhs->type && 
			VT_NUM == rhs->val.val.type && 
			ET_VALUE == lhs->type &&
			VT_NUM == lhs->val.val.type) {
		WRITE(rp, ep, RULE_ARITH);
		ep->type = ET_VALUE;
		ep->val.val.type = VT_NUM;
		ep->val.val.val.num = 
			lhs->val.val.val.num *
			rhs->val.val.val.num;
		expr_free(lhs);
		expr_free(rhs);
		return(1);
	}

	/* FIXME: fails! */
#if 0
	/* Multiplicative arithmetic: lhs->rhs. */

	if (ET_BINARY == rhs->type &&
			BT_DIV == rhs->val.bin.type) {
		WRITE(rp, ep, RULE_ARITH);
		memcpy(ep, rhs, sizeof(struct expr));
		free(rhs);
		rhs = ep->val.bin.lhs;
		ep->val.bin.lhs = expr_alloc(ET_BINARY);
		ep->val.bin.lhs->val.bin.type = BT_MUL;
		ep->val.bin.lhs->val.bin.lhs = lhs;
		ep->val.bin.lhs->val.bin.rhs = rhs;
		return(1);
	}

	/* Multiplicative arithmetic: rhs->lhs. */

	if (ET_BINARY == lhs->type &&
			BT_DIV == lhs->val.bin.type) {
		WRITE(rp, ep, RULE_ARITH);
		memcpy(ep, lhs, sizeof(struct expr));
		free(lhs);
		rhs = ep->val.bin.lhs;
		ep->val.bin.lhs = expr_alloc(ET_BINARY);
		ep->val.bin.lhs->val.bin.type = BT_MUL;
		ep->val.bin.lhs->val.bin.lhs = lhs;
		ep->val.bin.lhs->val.bin.rhs = rhs;
		return(1);
	}
#endif

	if (ET_VALUE == lhs->type && VT_NUM == lhs->val.val.type)
		if (expr_rdc_mulr(rp, lhs, rhs))
			return(1);

	if (ET_VALUE == rhs->type && VT_NUM == rhs->val.val.type)
		if (expr_rdc_mulr(rp, rhs, lhs))
			return(1);

	return(0);
}

/*
 * Recursively scan through binary multiplication and put together like
 * terms.
 */
static int
expr_rdc_mulr(const struct rdc *rp, 
		struct expr *ep1, struct expr *ep2)
{

	if (ET_VALUE == ep2->type) {
		if (VT_NUM != ep2->val.val.type)
			return(0);
		ep2->flags |= EXPR_RDC;
		WRITE(rp, ep1, RULE_ARITH);
		ep2->flags &= ~EXPR_RDC;
		ep1->val.val.val.num = 
			ep1->val.val.val.num *
			ep2->val.val.val.num ;
		ep2->val.val.val.num = 1.0L;
		return(1);
	} else if (ET_BINARY != ep2->type)
		return(0);

	assert(ET_BINARY == ep2->type);

	if (BT_MUL != ep2->val.bin.type)
		return(0);

	if (expr_rdc_mulr(rp, ep1, ep2->val.bin.lhs))
		return(1);
	if (expr_rdc_mulr(rp, ep1, ep2->val.bin.rhs))
		return(1);

	return(0);
}

static int
expr_rdc_addsub(const struct rdc *rp, struct expr *ep)
{
	struct expr	*lhs, *rhs;
	enum btype	 bt;

	bt = ep->val.bin.type;
	lhs = ep->val.bin.lhs;
	rhs = ep->val.bin.rhs;

	/* Normalise subtraction. */

	if (BT_SUB == bt) {
		WRITE(rp, ep, RULE_ARITH);
		ep->val.bin.type = BT_ADD;
		ep->val.bin.rhs = expr_alloc(ET_UNARY);
		ep->val.bin.rhs->val.un.type = UT_NEG;
		ep->val.bin.rhs->val.un.val = rhs;
		return(1);
	}

	/* Identity of addition/subtraction (lhs). */

	if (ET_VALUE == lhs->type && 
			VT_NUM == lhs->val.val.type && 
			0.0L == lhs->val.val.val.num) {
		WRITE(rp, ep, RULE_IDENT);
		expr_free(lhs);
		memcpy(ep, rhs, sizeof(struct expr));
		free(rhs);
		return(1);
	}

	/* Identity of addition/subtraction (rhs). */

	if (ET_VALUE == rhs->type && 
			VT_NUM == rhs->val.val.type && 
			0.0L == rhs->val.val.val.num) {
		WRITE(rp, ep, RULE_IDENT);
		expr_free(rhs);
		memcpy(ep, lhs, sizeof(struct expr));
		free(lhs);
		return(1);
	}

	/* Addition/subtraction arithmetic. */

	if (ET_VALUE == rhs->type && 
			VT_NUM == rhs->val.val.type && 
			ET_VALUE == lhs->type &&
			VT_NUM == lhs->val.val.type) {
		WRITE(rp, ep, RULE_ARITH);
		ep->type = ET_VALUE;
		ep->val.val.type = VT_NUM;
		ep->val.val.val.num = 
			BT_ADD == bt ?
			lhs->val.val.val.num +
			rhs->val.val.val.num :
			lhs->val.val.val.num -
			rhs->val.val.val.num;
		expr_free(lhs);
		expr_free(rhs);
		return(1);
	}

	if (BT_ADD != bt)
		return(0);

	if (ET_VALUE == lhs->type && VT_NUM == lhs->val.val.type)
		if (expr_rdc_addsubr(rp, lhs, rhs))
			return(1);

	if (ET_VALUE == rhs->type && VT_NUM == rhs->val.val.type)
		if (expr_rdc_addsubr(rp, rhs, lhs))
			return(1);

	return(0);
}

/*
 * Recursively scan through binary addition and put together like terms.
 * This only operates over addition because subtraction is normalised
 * out (requiring double-negation etc.).
 */
static int
expr_rdc_addsubr(const struct rdc *rp, 
		struct expr *ep1, struct expr *ep2)
{

	if (ET_VALUE == ep2->type) {
		if (VT_NUM != ep2->val.val.type)
			return(0);
		ep2->flags |= EXPR_RDC;
		WRITE(rp, ep1, RULE_ARITH);
		ep2->flags &= ~EXPR_RDC;
		ep1->val.val.val.num = 
			ep1->val.val.val.num +
			ep2->val.val.val.num ;
		ep2->val.val.val.num = 0.0L;
		return(1);
	} else if (ET_BINARY != ep2->type)
		return(0);

	assert(ET_BINARY == ep2->type);

	if (BT_ADD != ep2->val.bin.type)
		return(0);

	if (expr_rdc_addsubr(rp, ep1, ep2->val.bin.lhs))
		return(1);
	if (expr_rdc_addsubr(rp, ep1, ep2->val.bin.rhs))
		return(1);

	return(0);
}

static int
expr_rdc_neg(const struct rdc *rp, struct expr *ep)
{
	struct expr	*lhs, *rhs;
	enum btype	 bt;

	lhs = ep->val.un.val;

	/* Normalise negation. */

	if (ET_VALUE == lhs->type && 
			VT_NUM == lhs->val.val.type) {
		WRITE(rp, ep, RULE_ARITH);
		memcpy(ep, lhs, sizeof(struct expr));
		free(lhs);
		ep->val.val.val.num = -ep->val.val.val.num;
		return(1);
	}

	/* Normalise binary negation. */

	if (ET_BINARY == lhs->type && 
			(BT_ADD == lhs->val.bin.type ||
			 BT_SUB == lhs->val.bin.type)) {
		bt = lhs->val.bin.type;
		WRITE(rp, ep, RULE_ARITH);
		memcpy(ep, lhs, sizeof(struct expr));
		free(lhs);
		lhs = ep->val.bin.lhs;
		rhs = ep->val.bin.rhs;
		ep->val.bin.lhs = expr_alloc(ET_UNARY);
		ep->val.bin.lhs->val.un.type = UT_NEG;
		ep->val.bin.lhs->val.un.val = lhs;

		if (BT_SUB == bt) {
			ep->val.bin.rhs = rhs;
			return(1);
		}

		ep->val.bin.rhs = expr_alloc(ET_UNARY);
		ep->val.bin.rhs->val.un.type = UT_NEG;
		ep->val.bin.rhs->val.un.val = rhs;
		return(1);
	}

	/* Reduce: double-negation. */

	if (ET_UNARY == lhs->type &&
			UT_NEG == lhs->val.un.type) {
		WRITE(rp, ep, RULE_ARITH);
		memcpy(ep, lhs, sizeof(struct expr));
		free(lhs);
		lhs = ep->val.un.val;
		memcpy(ep, lhs, sizeof(struct expr));
		free(lhs);
		return(1);
	}

	return(0);
}

static int
expr_rdc_ln(const struct rdc *rp, struct expr *ep)
{
	struct expr	*lhs;
	struct sym	*e;

	lhs = ep->val.un.val;

	/* Properties of logarithm. */

	switch (lhs->type) {
	case (ET_BINARY):
		switch (lhs->val.bin.type) {
		case (BT_MUL):
			WRITE(rp, ep, RULE_PLN);
			ep->type = ET_BINARY;
			ep->val.bin.type = BT_ADD;
			ep->val.bin.lhs = expr_alloc(ET_UNARY);
			ep->val.bin.lhs->val.un.type = UT_LN;
			ep->val.bin.lhs->val.un.val = lhs->val.bin.lhs;
			ep->val.bin.rhs = expr_alloc(ET_UNARY);
			ep->val.bin.rhs->val.un.type = UT_LN;
			ep->val.bin.rhs->val.un.val = lhs->val.bin.rhs;
			free(lhs);
			return(1);
		case (BT_DIV):
			WRITE(rp, ep, RULE_PLN);
			ep->type = ET_BINARY;
			ep->val.bin.type = BT_SUB;
			ep->val.bin.lhs = expr_alloc(ET_UNARY);
			ep->val.bin.lhs->val.un.type = UT_LN;
			ep->val.bin.lhs->val.un.val = lhs->val.bin.lhs;
			ep->val.bin.rhs = expr_alloc(ET_UNARY);
			ep->val.bin.rhs->val.un.type = UT_LN;
			ep->val.bin.rhs->val.un.val = lhs->val.bin.rhs;
			free(lhs);
			return(1);
		case (BT_EXP):
			WRITE(rp, ep, RULE_PLN);
			ep->type = ET_BINARY;
			ep->val.bin.type = BT_MUL;
			ep->val.bin.lhs = lhs->val.bin.rhs;
			ep->val.bin.rhs = expr_alloc(ET_UNARY);
			ep->val.bin.rhs->val.un.type = UT_LN;
			ep->val.bin.rhs->val.un.val = lhs->val.bin.lhs;
			free(lhs);
			return(1);
		default:
			break;
		}
		break;
	case (ET_VALUE):
		switch (lhs->val.val.type) {
		case (VT_IDENT):
			e = symtab_add(rp->tab, "e", 1);
			if (e != lhs->val.val.val.sym)
				break;
			WRITE(rp, ep, RULE_PLN);
			ep->type = ET_VALUE;
			ep->val.val.type = VT_NUM;
			ep->val.val.val.num = 1.0L;
			expr_free(lhs);
			return(1);
		default:
			if (1.0L != lhs->val.val.val.num) 
				break;
			WRITE(rp, ep, RULE_PLN);
			ep->type = ET_VALUE;
			ep->val.val.type = VT_NUM;
			ep->val.val.val.num = 0.0L;
			expr_free(lhs);
			return(1);
		}
		break;
	default:
		break;
	}

	return(0);
}

static int
expr_rdc_un(const struct rdc *rp, struct expr *ep)
{

	expr_rdc(rp, ep->val.un.val);

	assert(*expr_un_rdcs[ep->val.un.type]);
	return((*expr_un_rdcs[ep->val.un.type])(rp, ep));
}

static int
expr_rdc_bin(const struct rdc *rp, struct expr *ep)
{

	expr_rdc(rp, ep->val.bin.lhs);
	expr_rdc(rp, ep->val.bin.rhs);

	switch (ep->val.bin.type) {
	case (BT_ADD):
		return(expr_rdc_addsub(rp, ep));
	case (BT_SUB):
		return(expr_rdc_addsub(rp, ep));
	case (BT_MUL):
		return(expr_rdc_mul(rp, ep));
	case (BT_EXP):
		return(expr_rdc_exp(rp, ep));
	case (BT_DIV):
		return(expr_rdc_div(rp, ep));
	default:
		break;
	}

	return(0);
}

void
expr_rdc(const struct rdc *rp, struct expr *ep)
{

	/*
	 * Repeatedly run the expression while there's something to
	 * reduce with it.  Our expression type can change, but all
	 * expressions have handlers so it's ok.
	 */

	assert(*expr_rdcs[ep->type]);
	while ((*expr_rdcs[ep->type])(rp, ep))
		assert(*expr_rdcs[ep->type]);
}

/*
 * Expression reduction applies mathematical laws to reduce an equation,
 * e.g., ln(e) = 1.  The aim is to remove superfluous mathematical
 * expressions from an equation.
 */
void
expr_reduce(struct writer *write, struct symtab *tab, 
		const struct expr *top, struct expr *ep)
{
	struct rdc	r;

	r.write = write;
	r.top = top;
	r.tab = tab;

	/* Reduce an expression, starting at a (sub-)expression. */
	expr_rdc(&r, ep);
}
